// Module.h : main header file for TModule and related classes.
//

#if !defined(AFX_MODBASE_H__199E2135_C73F_11D2_8045_00105A1E5D53__INCLUDED_)
#define AFX_MODBASE_H__199E2135_C73F_11D2_8045_00105A1E5D53__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#ifndef __AFXWIN_H__
#error include 'stdafx.h' before including this file for PCH
#endif

#include <afxmt.h>
#include <afxtempl.h>
#include "sysdef.h"
#include "Utils/Collections.h"
#include "Utils/Semaphore.h"
#include "Utils/Time.h"
#include "Utils/Users.h"
#include "Utils/Value.h"
#include "Xpert/Win32Fix.h"
#include "XmlSetup.h"

#undef DECL
#ifdef ENGINE_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif


///////////////////////////////////////////////////////////////////////////////
// Module base class and its supporting classes
//

/*
Diagram which attempts to show the various screen attributes used for displaying the graphical
sensor hookup screen:

--------------------------------------------------------------               ---
(tm)                                                                             |
--------------------------------------------------------------                |
(it)                                                        |
+-----------+                  +-----------+ (iom)      ---     |
(lm)          ---|           |---  (rm)      ---|           |---          |      |
(mi) --|           |-----+         ---|           |--- (mo)     |      |
+-------|           |---  |    (ios)---|(il)   (ir)|---        (ih)   (ch)
|    ---|           |---  |         ---|           |---          |      |
|    ---|           |---  |      +-----|           |---          |      |
|       +-----------+     |      |     +-----------+ (iom)      ---     |
(ib)          +------+     <-----iw---->                    |
|
--------------------------------------------------------------                |
(bm)                                                                             |
--------------------------------------------------------------               ---
<---------------(cw)-------------------->
*/


/*

Module Property Design:

1) Properties are stored in a variant data structure so we can support multiple data types,
every data type can be converted to and from strings so we can write out a setup or edit.

2) Each derivation of TModule can define it's own list of Properties and Variants.

TPropertyList MyList[] = { {"Slope", &Slope}, {"Offset", &Offset} };
void MyModule::SetupProperties()
{
MyAncestor::SetupProperties();
AddProperties(MyList);
};

3) AddProperties would add each property to a MapStringToPtr structure to allow for
fast lookup by name.

4) Finally we need some functions to enumerate all the available properties to
support streaming them out to disk.

Note: there may be some overlap between properties and real time values which can
be eliminated.

*/


//////////////////////////////////////////////////////////////////////////
// The following constants are used in Module base class implementation

//in order to have more than 5 outputs for engine execution purposes
//we define a seperate number of max outputs, for sdisplay purposes
//there is still a number limited by the size of icons
namespace ENG
{
   const int MaxOutputs = 20;
}

namespace UI
{
   const int MaxInputs                = 5;  // (mi)
   const int MaxOutputs               = 5;  // (mo)
   const int ZoomLevels               = 2;
};

namespace UIPrivate
{
   const int LabelHeight[UI::ZoomLevels]  = {10,10}; // Text label below the icon
   const int LeftMargin[UI::ZoomLevels]   = {5, 10};  // (lm
   const int RightMargin[UI::ZoomLevels]  = {5, 10};  // (mo)
   const int TopMargin[UI::ZoomLevels]    = {5, 10};  // (mtr)
   const int BottomMargin[UI::ZoomLevels] = {5 + LabelHeight[0], 5 + LabelHeight[1]};  // (mbr)
   const int IconHeight[UI::ZoomLevels]   = {32, 84}; // (ih)
   const int IconWidth[UI::ZoomLevels]    = {32, 84}; // (iw)
   const int IOSpacing[UI::ZoomLevels]    = {IconHeight[0]/UI::MaxInputs, IconHeight[1]/UI::MaxInputs};                    // (ios)
   const int IOMargin[UI::ZoomLevels]     = {IconHeight[0]%UI::MaxInputs/2, IconHeight[1]%UI::MaxInputs/2};    // (iom)
   const int IconTop[UI::ZoomLevels]      = {TopMargin[0],TopMargin[1]};             // (it)
   const int IconBottom[UI::ZoomLevels]   = {IconTop[0] + IconHeight[0] - 1, IconTop[1] + IconHeight[1] - 1};        // (ib)
   const int IconLeft[UI::ZoomLevels]     = {LeftMargin[0], LeftMargin[1]};          // (il)
   const int IconRight[UI::ZoomLevels]    = {IconLeft[0] + IconWidth[0] - 1,IconLeft[1] + IconWidth[1] - 1};                // (ir)
   const int CellWidth[UI::ZoomLevels]    = {IconWidth[0] + LeftMargin[0] + RightMargin[0], IconWidth[1] + LeftMargin[1] + RightMargin[1]};    // (cw)
   const int CellHeight[UI::ZoomLevels]   = {IconHeight[0] + TopMargin[0] + BottomMargin[0], IconHeight[1] + TopMargin[1] + BottomMargin[1]};   // (ch)
};


class DECL TProperty : public TValue
{
public:
   TProperty()                              {Attributes=0; };
   TProperty& operator=(const TProperty& o) {Attributes=o.Attributes; TValue::operator=(o); return *this;};
   TProperty& operator=(const TValue& o)    {TValue::operator=(o);  return *this;};
   TProperty& operator=(TCHAR* Str)         {TValue::operator=(Str); return *this;};
   TProperty& operator=(LPCTSTR Str)        {TValue::operator=(Str); return *this;};
   TProperty& operator=(const CString Str)  {TValue::operator=(Str); return *this;};
   TProperty& operator=(int Num)            {TValue::operator=(Num); return *this;};
   TProperty& operator=(bool Num)           {TValue::operator=(Num); return *this;};
   TProperty& operator=(double Num)         {TValue::operator=(Num); return *this;};
public:
   enum TAttributes {NOSAVE=1, READONLY=2, SYSTEM=4, DYNAMIC=8};
   int Attributes;
};

///////////////////////////////////////////////////////////////////////////
// TPropertyList class
// 
class DECL TPropertyList : public TMapStringToRef<TProperty>
{
public:
   TPropertyList() {};
   virtual ~TPropertyList();
   void Add(const CString Str, TProperty& Property)    { SetAt(Str, Property); };
   void Add(const CString Str, TProperty& Property, TProperty::TAttributes Attr)  { Property.Attributes=Attr; Add(Str, Property); };
   bool Remove(const CString Str)                    { return RemoveKey(Str) != 0; };
   bool Get(const CString Str, TProperty*& Property) { return Lookup(Str, Property) != 0; };
   bool GetNext(POSITION& P, CString& Str, TProperty*& Property) const { if (P)  { GetNextAssoc(P, Str, Property); return true; } else return false; };
   
   bool ReadSetup(CXMLSetup& XMLSetup);   // Read setup
   bool WriteSetup(CXMLSetup& XMLSetup);  // Writes setup, clears changed flags
};

/////////////////////////////////////////////////////////////////////////////
// TPropertyContainer class
//
class DECL TPropertyContainer : public CObject
{
   // Interface
public:
   TPropertyContainer() {}
   virtual ~TPropertyContainer() {};
   void AddProperty(TProperty& prop);
   void AddProperty(const CString& str, TProperty& prop);
   bool RemoveProperty(const CString& str) {return m_props.Remove(str);}
   bool GetProperty(const CString str, TProperty*& prop) {return m_props.Get(str, prop);}
   int GetPropertyCount() {return m_props.GetCount();}

   // Setup methods. 
   virtual CString GetSetupTag() = 0;
   virtual bool InitSetup() {return true;}
   virtual bool ReadSetup(CXMLSetup& xmlSetup) {return m_props.ReadSetup(xmlSetup);}
   virtual bool WriteSetup(CXMLSetup& xmlSetup);
   virtual void AfterReadSetup() {};
   virtual void BeforeWriteSetup() {};

   // Implementation
   protected:
   friend class TPropertyContainerEnum;
   TPropertyList m_props;
};

/////////////////////////////////////////////////////////////////////////////
// TPropertyEnum class
//
//    Example usage: 
// 
//       TPropertyEnum propEnum(container);
//       TProperty* pProp;
//       while (propEnum.Get(pProp))
//          ... use property ...
//
class TPropertyContainerEnum
{
   // Interface
public:
   TPropertyContainerEnum(TPropertyContainer& container) {list = &container.m_props; p = list->GetStartPosition();}
   bool Get(CString& str, TProperty*& prop) {return list->GetNext(p, str, prop);}
   bool Get(TProperty*& prop) {CString str; return list->GetNext(p, str, prop);}

   // Implementation   
   protected:
   TPropertyList* list;
   POSITION p;
};

///////////////////////////////////////////////////////////////////////////////
// CModPropPage - wrapper class used by modules that create property pages.
//
class DECL CModPropPage : public CPropertyPage
{
public:
   // Constructors
   CModPropPage(UINT nIDTemplate, UINT nIDCaption = 0)
      : CPropertyPage(nIDTemplate, nIDCaption) { }
   CModPropPage(LPCTSTR lpszTemplateName, UINT nIDCaption = 0)
      : CPropertyPage(lpszTemplateName, nIDCaption) { }

   virtual bool IsEnabledForDisplay() {return true;}

   // Attributes
   TUserGroup m_ug;           // The group this module is in
   TUserGroup LoginGroup;   // The group user logged in as
};


///////////////////////////////////////////////////////////////////////////////
// TNode represents a connection point between modules.

//forward declare
class TModule;
typedef TModule * PModule;


//a node basically manages the connection point and a ptr to the module connected
class DECL TNode : public CObject
{
public:
   TNode()                                { Clear();};
   TNode(TModule& AModule, int APoint)    { Module=&AModule; point=APoint;      };
   TNode(const TNode& n)                  { Module=n.Module; point=n.point;     };
   virtual ~TNode()                       {};
   bool operator==(const TNode& n) const  { return ((n.Module==Module) && (n.point==point)); };
   bool operator!=(const TNode& n) const  { return ((n.Module!=Module) || (n.point!=point)); };
   TNode& operator=(const TNode& n)       { Module=n.Module; point=n.point; return *this;    };
public:
   void Clear()                           { Module=NULL; point=-1;              };
   bool Empty()                           { return Module==NULL;                };
   bool Inuse()                           { return Module!=NULL;                };
   TNode& Source();            // returns node which is the source of this node
   bool IsInput()                         { return (point >= 0) && (point < UI::MaxInputs);  };
   bool IsOutput()                        { return (point >= 0) && (point >= UI::MaxInputs);  };
   void SetInput(int Pt)                  { point = Pt;                         };
   void SetOutput(int Pt)                 { point = Pt + UI::MaxInputs;         };

   //Modified for "more" output capability, Do we want to keep this?
   //there are 2 types of points now, one for display of outputs in icon and
   //another for execution of engine.
   int Point();//  { return (point < UI::MaxInputs) ? point : point-UI::MaxInputs; };
   int EngPoint();

public:
   PModule Module;
private:
   int point;
};


/////////////////////////////////////////////////////////////////////////////
// CSensorData, used to pass data among modules

#define realtype double

class DECL CSensorData : public CObject
{
public:
   virtual ~CSensorData();
   CSensorData(const CSensorData& data);
   CSensorData() { Reset(); };
   void Reset(); 
   BOOL operator == (const CSensorData& data) const;
   CSensorData& operator=(const CSensorData& data);
   void SetID(CString& string){SensorID = string;};
   CString& GetID(){return SensorID;};
   CString FormatData();
   CString FormatData(int iRightDigits);
   
   // The basic quality types Xpert supports. 
   enum QualityType {GOOD, BAD, UNDEFINED};

   // This tests quality based on whether it's not known bad, as opposed to known 
   // good, which can be useful when custom defined quality flags are in use (e.g., LEADS). 
   virtual bool IsGood() {return Quality != BAD && Quality != UNDEFINED;}

   // Classes can derive from CSensorData to implement special quality flag 
   // handling. By default, CSensorData treats unrecognized values as "undefined".
   virtual CString GetQualStr();
   virtual void SetQualStr(const CString& strQual);
   
   // Classes deriving from CSensorData may need to define their own static 
   // conversion routines. 
   static CString QualityToStr(CSensorData::QualityType q);
   static QualityType StrToQuality(const CString& strQual);
   static CString AlarmToStr(DWORD dwAlarmStatus);

// 9000 SSP Alarm Status
   enum TAlarmStatus  // (* Controls how alarms are handled, and current status *)
   {
// Alarm Status Word
      HiLimitA        = 1<<0,
      LowLimitA       = 1<<1,
      IntHiLimitA     = 1<<2,
      IntLowLimitA    = 1<<3,
      ControlTimeoutA = 1<<4,
      ControlFailA    = 1<<5,
      ControlSoftA    = 1<<6, 
      ControlHardA    = 1<<7,
      DigitalA        = 1<<8,
      InhibitA        = 1<<9,
      ROCLimitA       = DigitalA,
// Change Status Word
      HiLimitC        = 1<<16,
      LowLimitC       = 1<<17,
      IntHiLimitC     = 1<<18,
      IntLowLimitC    = 1<<19,
      ControlTimeoutC = 1<<20,
      ControlFailC    = 1<<21,
      ControlSoftC    = 1<<22, 
      ControlHardC    = 1<<23,
      DigitalC        = 1<<24,
      InhibitC        = 1<<25,
      ROCLimitC       = DigitalC
   };

   // Overrides
   // ClassWizard generated virtual function overrides
   //{{AFX_VIRTUAL(CSensorData)
   //}}AFX_VIRTUAL

   // Implementation
public:
   TTime           TimeScheduled; //scheduled timestamp
   TTime           TimeActual;    //actual timestamp
   CString         SensorID;      //id of sensor supplying this data
   TValue          Data;          //data itself
   CString         Units;         //how to track units??
   QualityType     Quality;       //what is quality of this data
   DWORD           AlarmStatus;   //Alarm status of this value (bit fields)
   UINT32          DeviceTime;
};


enum CIODeviceType {UNKNOWN_DEVICE, ANALOG, DIGITAL, DISPLAY, SDI12, RS232_DEV, RS485_DEV, BASIC_SENSOR};

//////////////////////////////////////////////////////////////////////////////////
// TModule, the module base class that all other modules should be derived from
class CXMLSetup;        //fwd declare
class CXMLToken;
class CScheduleDesc;

class DECL TModule : public CObject
{
   friend class TPropertyEnum;
public:
   TModule(const CString& ModuleName, HINSTANCE hModInst);
   virtual ~TModule();
   ////////////////////////////////////////////////////////////////////////
   //The following functions are used for managing the GUI parts of modules

   int FindInput(TModule& Module);              // Returns Inputs[] index if Module is an input else -1.
   int FindInput(TNode Node);                   // Returns the Inputs[] index of an input connected to Node else -1
   int WiredInputCount();                       // Returns the number of inputs which are wired up.
   void RemoveInput(TModule& Module);           // Removes any inputs matching Module
   BOOL IsPtInUse(int nPt, TModule* pModule);
   int GetConnectionPoint(TModule* pSrcMod);
   int FindModuleIndex(TModule* pMod, TObArray<TModule>& ModList);

   // Virtual function that may be overridden if desired
   virtual int GetDefaultInput(bool GetUnused=false, int Index=0);  // Pick a default input for a just created module
   virtual int GetDefaultOutput(int Index=0);       // Pick a default output for a just created module
   virtual bool ShowProperties(CWnd* pParent) {return false;};   // Show edit properties dialog. Return true on change.
   virtual int Rename(CWnd* pParent);               // Get a new function name from user.
   virtual bool GetIOAddrAsString(CString& string); //get a string describing I/O location, return false if unassigned
   virtual bool GetSensorInfo(CArray<CString, CString&>& info); //get strings to display in View Sensors, return false if no data should be displayed
   virtual CString OutputNamePopup(int Point){return OutputName(Point);};
   virtual CBitmap& GetBitmap(CBitmap& bm);     // Loads & returns a 32x32 bitmap to represent the module
   virtual void OnWireChange(CWnd* pParent) {}; // Called when change to module wiring occurs.
   virtual void AssignDefaultChannel() {};      // Called to setup default wiring for a new module.
   virtual void ClearChannelsInuse()  {};       // Should clear any I/O channel assignments
   virtual void SetChannelsInuse()    {};       // Should make any necessary I/O channel assignments
   virtual CIODeviceType GetDeviceType()          // Returns the type of device assuming it's an I/O module
      {return UNKNOWN_DEVICE;};
   virtual TProperty* GetModuleProperty()       // Returns the property that controls the I/O module number, Port, or Address
      {return NULL;};
   virtual TProperty* GetChannelProperty()      // Returns the property that controls the channel number
      {return NULL;};
   virtual TProperty* GetSlopeProperty(int Point);  // Returns the slope property associated with an output point
   virtual TProperty* GetOffsetProperty(int Point); // Returns the offset property associated with an output point
   virtual TProperty* GetMeasurementIntervalProperty(int iPoint = -1); // Returns the measurement interval property associated with an output point
   virtual TProperty* GetMeasurementTimeProperty(int iPoint = -1); // Returns the measurement time property associated with an output point
   virtual TProperty* GetSampleIntervalProperty(int iPoint = -1); // Returns the sample interval property associated with an output point
   virtual TProperty* GetSampleDurationProperty(int iPoint = -1); // Returns the sample duration property associated with an output point

   //pure virtuals that must be defined for each module
   virtual CString InputName(int Point) = 0;    // Returns the name of an Input
   virtual CString OutputName(int Point) = 0;   // Returns the name of an Output
   virtual bool InputActive(int Point) = 0;     // Returns true if a module uses an input point
   virtual bool OutputActive(int Point) = 0;    // Returns true if a module uses an output point
   virtual TModule* CreateAnother() = 0;        // Calls default constructor to create another module of the same type

   //////////////////////////////////////////////////////////////////////////////
   //These functions are for support of reading/writing modules into setup files
protected:
   TPropertyList PropertyList;

public:
   void AddProperty(const CString Str, TProperty& Property)  { PropertyList.Add(Str, Property);         };
   void AddProperty(const CString Str, TProperty& Property, TProperty::TAttributes Attr)  { PropertyList.Add(Str, Property, Attr); };
   bool RemoveProperty(const CString Str)                    { return PropertyList.Remove(Str); };
   bool GetProperty(const CString Str, TProperty*& Property) { return PropertyList.Get(Str, Property); };

   // Override to prohibit dynamic property creation.
   virtual bool SupportsDynamicProperties(void) { return true; };

public:
   void WriteXMLSetup(CXMLSetup& XMLSetup, TObArray<TModule>& ModuleList, BOOL fDefaults);
   BOOL ReadXMLSetup(CXMLSetup& XMLSetup, CXMLToken& XMLToken, TObArray<TModule>& ModuleList, int BaseIndex);

   //virtual functions that may be overridden
   virtual void AfterReadXMLSetup(int BaseIndex);
   virtual void AfterReadSetup(int BaseIndex);
   virtual void BeforeWriteSetup();
   virtual bool IsActive() { return false; };            // True when block initiates data flow (e.g., Measure).
   virtual CString GetScheduleInfo() { return _T(""); }; // Return textual description of schedule (when module is active).

public: 
   // Look for an enabled active block in-line that will eventually execute us
   // Lock setup before calling when calling from a context where ModuleList could change. 
   virtual bool IsEnabled();
   virtual bool IsEnabledUpstream();
   virtual bool IsEnabledDownstream();

   // Determine time of next execution by examining connected blocks
   // Lock setup before calling when calling from a context where ModuleList could change. 
   virtual TTime GetNextExecuteTime();
   virtual TTime GetNextExecuteTimeUpstream();
   virtual TTime GetNextExecuteTimeDownstream();

   // Retrieve schedule most closely associated with this block by examining connected blocks
   // Lock setup before calling when calling from a context where ModuleList could change. 
   virtual bool GetSchedule(CScheduleDesc& schedDesc);
   virtual bool GetScheduleUpstream(CScheduleDesc& schedDesc);
   virtual bool GetScheduleDownstream(CScheduleDesc& schedDesc);

   // Find logging details from nearest log block, favoring upstream. 
   // Lock setup before calling when calling from a context where ModuleList could change. 
   // NULL pMod means "this"
   bool FindLogBlock(CString &strLogFile, CString &strSensorName, TModule* pMod = NULL);
   bool FindLogBlockUpstream(CString &strLogFile, CString &strSensorName, TModule* pMod = NULL);
   bool FindLogBlockDownstream(CString &strLogFile, CString &strSensorName, TModule* pMod = NULL);

   // Find nearest property with the given label in connected blocks, favoring upstream. 
   // Lock setup before calling when calling from a context where ModuleList could change. 
   TProperty* FindProperty(LPCTSTR szName, TModule* pMod = NULL); // NULL pMod means "this"
   TProperty* FindPropertyUpstream(LPCTSTR szName, TModule* pMod = NULL);
   TProperty* FindPropertyDownstream(LPCTSTR szName, TModule* pMod = NULL);

public: // GUI related publics that need to be set
   virtual int InputCount() = 0;                // Number of active inputs & outputs
   virtual int OutputCount() = 0;               // Number of active outputs
   CString Name;                                // Name of the Module. (i.e. "Measure")

public: // GUI related publics that are calculated.
   int X, Y;                                    // Position of ICON on the screen
   bool bSelected;                              // GUI icon is selected for display.

public:
   //This node array tracks the input connections to this module
   TNode Inputs[UI::MaxInputs];

   //This index array tracks inputs but without the Module ptr.
   //Needed for saving the adjacent modules in the setup file
   int Indices[UI::MaxInputs];

   /////////////////////////////////////////////////////////////////////////////////
   //additions for support of run-time engine.

   //list of i2c event types
   enum I2CEvent {ALARM = 0x01, OVERSPEED = 0x02};

   //virtual functions that may be overridden
   virtual void Initialize() {};                            // to prepare for running engine
   bool Push(TTime tScheduled);                             // Execute and push the result to the outputs
   bool Pull(TTime tScheduled, bool bLiveReading);          // Execute inputs and compute new result. Set bLiveReading to true to override InhibitPull logic.
   void Get(int Point, CSensorData& Data);                  // Get Last Measured Value for a Point
   void Set(int Point, const CSensorData& Data);            // Set the Last Measured Value for a Point
   bool GetInputData(int Point, CSensorData& Data);         // Get Last Measure Value from an Input Module
   bool SetInputData(int Point, CSensorData& Data);         // Set Last Measure Value from an Input Module
   void KickBack(CSensorData& Data);                        // Kick a modified value back to anyone interested in it

   virtual int GetModuleType() {return TModule::INPUT;};    // what type of module is this (input, telemetry etc)
   virtual void Stop() {};                                  // for handling any cleanup on stop
   virtual void EventExec(BYTE EventType, UINT32 Time) {};
   virtual int GetIndexOfDataToCalibrate();                 // Override to define which output can be calibrated (defaults to default output)
   virtual bool SupportsGUICalibrate();                     // Override if there is a GUI-based cal routine
   virtual bool GUICalibrate(void);                         // Override with actual calibration routine. It should return true to force setup save.
   virtual bool SupportsI2CCalibrate(void);                 // Override if there is a i2c LCD-based cal routine
   virtual bool I2CCalibrate(void);                         // Override with actual calibration routine. It should return true to force setup save.
   virtual bool Calibrate(CSensorData& sdOld, CSensorData& sdNew); // Override if needed to apply cal in custom fashion. Return true to save setup. 
   virtual void ExecuteScheduledAction(TTime tScheduled, bool bEvent) { Pull(tScheduled, bEvent); };   // Hook to provide action for schedule object.
   virtual bool InhibitPull(TTime tScheduled);              // Hook to define when Pull should be inhibited. Ignored if a force is in effect.
   virtual void ClearAlarm() {};                            // Override to define what should happen on Engine.ClearAlarm()
   //list of module types
   enum ModuleType {INPUT, OUTPUT, PROCESSING, CALCULATION, LOGGING, TELEMETRY, OTHER};
   TObArray<TNode> Outputs;

protected: // Internal implementations of Public functions:
   virtual void Execute(TTime tScheduled) {};               // Compute current values
   virtual void Execute(TTime tScheduled, bool bLiveReading) { Execute(tScheduled); }; // Compute current values but make bLiveReading flag available
   virtual void PushData(TTime tScheduled);                                            // An event is pushing data
   virtual void PullData(TTime tScheduled, bool bLiveReading);                         // The module is being polled for data
   virtual void SetData(int Point, const CSensorData& Data)  { LastData = Data; };     // Set the Last Measured Value for a Point
   virtual void GetData(int Point, CSensorData& Data)        { Data = LastData; };     // Get Last Measured Value for a Point
   virtual void OnKickBack(CSensorData& Data);               // Send the data back to the block on input 2, blocks can store the modified value if they wish by over-riding
   virtual DWORD GetLiveReadingTimeout()                     { return 5000;     };     // How long to wait before aborting a livereading (0 means LiveReadings shouldn't be taken)

protected:
   CSensorData LastData;   // Default LastData

public: // Status information
   TTime tLastExecute;     // Absolute time of last execute.
   TTime tLastScheduled;   // Time of last scheduled execution (set by CSchedule)
   TTime tNextScheduled;   // Time of next scheduled execution (set by CSchedule)
   TTime tFirstSample;     // Time of the first sample to take (set by CSchedule)

protected: // Helper functions
// bool GetInputData(int Point, CSensorData& Data);               // Get Last Measure Value from an Input Module
   bool PullInputs(TTime tScheduled, bool bLiveReading);          // Pull Data from all Inputs
   bool PushOutputs(TTime tScheduled);                            // Push Data to all outputs
   bool PushOutputs(TTime tScheduled, int Output);                // Push Data to a specific output
   void LockData();                                               // Prevent access to Data
   void UnlockData();                                             // Allow access to Data

protected:
   TSemaphore PullLock;                            // for thread-safe access of Pull's
   TSemaphore PushLock;                            // for thread-safe access of Push'sll's
   CCriticalSection DataLock;                      // For GetData(), SetData() calls
   HINSTANCE ModuleInstance;

public: // PROPERTIES
   TProperty Function;                             // GUI block label
   TProperty Input[UI::MaxInputs];
   TProperty Index;
   TProperty IsGUI;                                // True for modules editable with the setup GUI
};


// Help class for enumerating property lists.
class TPropertyEnum
{
public:
   TPropertyEnum(TModule& Module)                   { List = &Module.PropertyList; P = List->GetStartPosition(); };
   bool Get(CString& Str, TProperty*& Property)     { return List->GetNext(P, Str, Property);                    };
private:
   TPropertyList* List;
   POSITION P;
};


///////////////////////////////////////////////////////////////////////////
//Common Sutron DLL interface class
//
//This is a class which creates and describes TModule's and their PropertyPages.
class DECL TModuleFactory : public CObject
{
public:
   TModuleFactory()                               { ModuleCount = 0; ModuleList = NULL; PageCount = 0; PageList = NULL; };
   virtual ~TModuleFactory()                      {};
public: // Every DLL should over-ride the following
   virtual TModule* CreateModule(int Index) = 0;
   virtual CModPropPage* CreatePage(int Index) = 0;
public: // Only need to be over-ridden if the DLL's constructor doesn't initalize ModuleCount, ModuleList, PageCount, PageList
   virtual int GetModuleCount()                   { return ModuleCount;                    };
   virtual CString GetModuleName(int Index)       { return (CString) ModuleList[Index];    };
   virtual int GetPageCount()                     { return PageCount;                      };
   virtual CString GetPageName(int Index)         { return (CString) PageList[Index];      };
public:
   virtual int FindModule(const CString& Name);
   int FindPage(const CString& Name);
   TModule* CreateModule(const CString& Name)     { return CreateModule(FindModule(Name)); };
   CModPropPage* CreatePage(const CString& Name) { return CreatePage(FindPage(Name));     };
protected:
   int ModuleCount;
   int PageCount;
   TCHAR** ModuleList;
   TCHAR** PageList;
};


//This is a base class for modules that need to add an entry to
//the NextGen Control Panel tree
class DECL CControlPanelEntry : public CObject
{
public:
   CControlPanelEntry() {m_pTree = NULL; m_hRootItem = NULL; m_bRecStopped = false; m_bSaveNeeded = false;};
   virtual void InsertItems(CTreeCtrl * pTree) {};
   virtual bool EditItem(CWnd* pParent, HTREEITEM hItem, CTreeCtrl* pTree) {return false;};
   virtual bool DeleteItem(CWnd* pParent, HTREEITEM hItem, CTreeCtrl* pTree) {return false;};
   virtual bool NewItem(CWnd* pParent, HTREEITEM hItem, CTreeCtrl* pTree) {return false;};
   virtual void GetButtonLabels(HTREEITEM hItem, CString& NewButton, CString& EditButton, CString& DeleteButton) {};
   virtual void OnKillActive(CWnd* pParent) {};
   virtual bool PromptStopEngine();
   CTreeCtrl*  m_pTree;
   HTREEITEM m_hRootItem;
   bool m_bRecStopped;
   bool m_bSaveNeeded;
};


class DECL TDialog : public CDialog
{
public:
   TDialog() : CDialog() {};
   TDialog(LPCTSTR lpszTemplateName, CWnd* pParentWnd = NULL) : CDialog(lpszTemplateName, pParentWnd) {};
   TDialog(UINT nIDTemplate, CWnd* pParentWnd = NULL) : CDialog(nIDTemplate, pParentWnd) {};
public:
   virtual LRESULT WindowProc(UINT message, WPARAM wParam, LPARAM lParam );
private:
   bool TimerKludgeFix;
};


////////////////////////////////////////////////////////////////////////////
// Global utility functions, called by indiviual modules

// SetTimeDialog.cpp
int DECL SetTimeDlg(CString& text, CWnd * pParentWnd, const CString& title, 
      bool bMSec = false, bool b23HrLimit = false);

// SetDateTimeDlg.cpp
int DECL SetDateTimeDlg(CWnd * pParentWnd, TTime& time, SYSTEMTIME& SystemTime, 
                        LPCTSTR pstrTitle = NULL);

// FileNameDlg.cpp
int DECL FileNameDlg(CWnd * pParentWnd, CString& sReturn, LPCTSTR lpszDir = _T("\\"), 
      LPCTSTR lpszFilters = _T("*.*"), LPCTSTR lpszDefaultExt = _T(""), 
      LPCTSTR lpszDefaultName = NULL);

// QwertyKeypad.cpp
void DECL MakeNoise();
int DECL KeypadDlg(CWnd* pParent, CString& text, CString Caption = _T(""), CString strType = _T(""));
int DECL PasswordDlg(CWnd* pParent, CString& text);

// NumberDlg.cpp
int DECL ChangeNumberDlgReal(CWnd* pParent, realtype& rInput, LPCTSTR pszTitle = NULL);
int DECL ChangeNumberDlgInt(CWnd* pParent, int& nInput, LPCTSTR pszTitle = NULL);

// EditUnitsDlg.cpp
void DECL ChangeUnitsDlg(CWnd* pParent, CString& sUnits, BOOL& bUseNewUnits);

// MessageDlg.cpp. 
// This dialog is typically used to display a message for a limited amount of time. 
// Types: MB_OK, MB_OKCANCEL, MB_YESNO, and MB_YESNOCANCEL.
// A wait time of 0 causes dialog to wait until user presses a button. Otherwise, the 
// wait time is allowed to pass before forcing nDefaultResponse.
// Possible values for return and nDefaultResponse: IDOK, IDCANCEL, IDYES, and IDNO.
// MessageDlgCallback should not be called directly except by CGUIView.
int DECL MessageDlg(CWnd* pParent, const LPCTSTR tstrMsg, UINT nWaitSec = 0, 
      UINT nType = MB_OK, UINT nDefaultResponse = NULL);
int DECL MessageDlgCallback(CWnd* pParent, const LPCTSTR tstrMsg, UINT nWaitSec = 0, 
      UINT nType = MB_OK, UINT nDefaultResponse = NULL);

// MsgBox works just much like AfxMessageBox except it allows the caption to be set,
// and it automatically hides and restores the wait cursor.
int DECL MsgBox(LPCTSTR Prompt, UINT nType, LPCTSTR Caption);
// MessageBoxCallback should not be called directly except by CGUIView.
int DECL MessageBoxCallback(CWnd* pParent, const LPCTSTR tstrPrompt, 
      UINT nType, const LPCTSTR tstrCaption);

// NetworkPropertiesDlg.cpp
// This dialog displays a network properties dialog that allows the passed network
// settings to be view/changed, but it does not actually modify the settings in the
// system - that task is left to the caller.
int DECL NetworkPropertiesDlg(bool& UseDHCP, CString& IPAddress, CString& SubnetMask, CString& DefaultGateway, 
                              CString& PrimaryDNS,  CString& SecondaryDNS,
                              CString& PrimaryWINS, CString& SecondaryWINS);

// StatsDlg.cpp
// This dialog shows a text string with title in a full screen scrollable dialog, 
// with option to edit. When editable, buttons prompt "Ok/Cancel", else "Close"
int DECL StatsDlg(CWnd* pParent, CString& strStats, LPCTSTR lpszTitle, bool bReadOnly = true);

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_MODBASE_H__199E2135_C73F_11D2_8045_00105A1E5D53__INCLUDED_)
